#include "FastLED.h"

#define NUM_RING_LEDS 48
#define LED_RING_PIN A0
#define NUM_SCORE_LEDS 8
#define SCORE_PIN A5
#define BRIGHTNESS 50
#define BUTTON_PIN 4
#define BUTTON_LED 3
#define BUZZER_PIN 9

#define ST_GAME_START 0
#define ST_WIN 98
#define ST_LOSE 99

#define NR_ROUNDS 8
#define NR_FIRST_EZ_ROUNDS 4

CRGB leds[NUM_RING_LEDS];
CRGB sleds[NUM_SCORE_LEDS];

byte game_state = ST_GAME_START;
byte light_pos = 0;
byte level = ST_GAME_START;
bool is_big_interval = 0;
bool button_led_state = LOW;
bool button_led_is_active = true;

const byte led_speed[8] = {60, 50, 40, 35, 30, 25, 20, 15};
const CRGB score_colors[8] = {CRGB::Green, CRGB::ForestGreen, CRGB::Cyan, CRGB::Yellow, CRGB::OrangeRed, CRGB::Red, CRGB::Blue, CRGB::Magenta};

unsigned long time_since_last_led_change = 0;
bool find_random = false;
byte target = 0;

void setup() {
  FastLED.addLeds<WS2812B, LED_RING_PIN, GRB>(leds, NUM_RING_LEDS);
  FastLED.addLeds<WS2812B, SCORE_PIN, GRB>(sleds, NUM_SCORE_LEDS);
  pinMode(BUTTON_PIN, INPUT_PULLUP);
  pinMode(BUTTON_LED, OUTPUT);

  // Disable global interrupts
  cli();  
  configure_timer1();
  init_timer1();
  // Enable global interrupts
  sei();
  
  Serial.begin(9600);
  Serial.println("Reset");
}

void loop() {
  // put your main code here, to run repeatedly:
  FastLED.setBrightness(BRIGHTNESS);
  if (game_state == ST_GAME_START) {
    start_game();
  } else if (game_state <= NR_ROUNDS) {
    play_round();
  } else if (game_state == ST_WIN) {
    win();
  } else if (game_state == ST_LOSE) {
    lose();
  }
}

ISR(TIMER1_COMPA_vect) {
  // Blink button led
  if (button_led_is_active) {
    button_led_state = !button_led_state;
    digitalWrite(BUTTON_LED, button_led_state);
  }
}

void configure_timer1() {
  TCCR1A = 0;
  TCCR1B = 0;
  TCNT1 = 0;
  OCR1A = 31249;            // compare match register 16MHz/256/2Hz-1/0.5sec
  TCCR1B |= (1 << WGM12);   // CTC mode
  TCCR1B |= (1 << CS12);    // 256 prescaler 
}

void init_timer1() {
  TIMSK1 |= (1 << OCIE1A);  // enable timer compare interrupt
}

void start_game(void) {
  // Rotating rainbow effect
  uint8_t thisHue = beat8(30, 255);
    fill_rainbow(leds, NUM_RING_LEDS, thisHue, 255 / NUM_RING_LEDS);
    fill_rainbow(sleds, NUM_SCORE_LEDS, thisHue, 255 / NUM_SCORE_LEDS);

    if (digitalRead(BUTTON_PIN) == LOW) {
      button_led_is_active = false;
      digitalWrite(BUTTON_LED, LOW);
      
      light_pos = 0;
      find_random = true;
      delay(500);

      // Clear ring LEDs + descending sound
      for (byte i = 0; i < NUM_RING_LEDS; ++i) {
        leds[i] = CRGB::Black;
        delay(40);
        FastLED.show();
       int thisPitch = map (i, NUM_RING_LEDS, 0, 100, 1500);
       tone(BUZZER_PIN, thisPitch, 120);
      }

      // Clear score LEDs
      for (byte i = 0; i < NUM_SCORE_LEDS; ++i) {
        sleds[i] = CRGB::Black;
        delay(100);
        FastLED.show();
      }
      game_state = 1;
      time_since_last_led_change = 0;
      
      button_led_is_active = true;
    }
    FastLED.show();
}

void play_round(void) {
  int period = led_speed[game_state - 1];
  if (millis() > time_since_last_led_change + period) {
    time_since_last_led_change = millis();

    // Find a target randomly
    if (find_random) {
      target = random(42) + 3;
      find_random = false;
    }
      
    leds[target] = CRGB::Green;

    // If round is easy, make the interval bigger
    if (game_state <= NR_FIRST_EZ_ROUNDS) {
      is_big_interval = 1;
      leds[target - 1].setRGB(255, 110, 0);
      leds[target + 1].setRGB(255, 110, 0);
    } else {
      is_big_interval = 0;
    }
      
    sleds[game_state - 1] = score_colors[game_state - 1];
    move_light(target - is_big_interval, target + is_big_interval);
  }
  
  if (digitalRead(BUTTON_PIN) == LOW) {
    delay(300);
    // find_random = false;
    if (light_pos - 1 >= target - is_big_interval && light_pos - 1 <= target + is_big_interval) {
      level = game_state;
      game_state = 98;
    } else {
      game_state = 99;
    }
  }
}

void move_light(byte bound1, byte bound2) {
  leds[light_pos] = CRGB::Red;

  // Clear last position
  if (light_pos != 0 && (light_pos < bound1 + 1 || light_pos > bound2 + 1)) {
    leds[light_pos - 1] = CRGB::Black;
  }

  FastLED.show();
  ++light_pos;
  
  // Loop back
  if (light_pos >= NUM_RING_LEDS) {
    leds[light_pos - 1] = CRGB::Black;
    light_pos = 0;
  }
}

void pulse_three_times(CRGB color, int pitch) {
  button_led_is_active = false;
  digitalWrite(BUTTON_LED, LOW);
  
   for (byte i = 0; i < 3; ++i) {
    for (byte j = 0; j < NUM_RING_LEDS; ++j) {
      leds[j] = color;
        tone(BUZZER_PIN, pitch, 250);
    }
    FastLED.show();
    delay(500);
    clear_ring_leds();
    FastLED.show();
    delay(500);
  }
  button_led_is_active = true;
}

void win() {
  pulse_three_times(CRGB::Green, 1000);
  find_random = true;
  light_pos = 0;

  game_state = level + 1;
  if (game_state > NR_ROUNDS) {
    game_state = 0;
    win_all();
  }
}

void lose() {
  pulse_three_times(CRGB::Red, 200);
  game_state = 0;
}

void clear_ring_leds() {
  for (byte i = 0; i < NUM_RING_LEDS; ++i) {
    leds[i] = CRGB::Black;
  }
}
void win_all(){
  
}
